# -*- coding: utf-8 -*-
# Copyright (C) 2017 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
from IPy import IP

from ..error import ConfigError
from ..setting import undefined
from ..i18n import _
from .option import Option


class IPOption(Option):
    "represents the choice of an ip"
    __slots__ = tuple()
    _display_name = _('IP')

    def __init__(self, name, doc, default=None, default_multi=None,
                 requires=None, multi=False, callback=None,
                 callback_params=None, validator=None, validator_params=None,
                 properties=None, private_only=False, allow_reserved=False,
                 warnings_only=False):
        extra = {'_private_only': private_only,
                 '_allow_reserved': allow_reserved}
        super(IPOption, self).__init__(name, doc, default=default,
                                       default_multi=default_multi,
                                       callback=callback,
                                       callback_params=callback_params,
                                       requires=requires,
                                       multi=multi,
                                       validator=validator,
                                       validator_params=validator_params,
                                       properties=properties,
                                       warnings_only=warnings_only,
                                       extra=extra)

    def _validate(self, value, context=undefined, current_opt=undefined):
        # sometimes an ip term starts with a zero
        # but this does not fit in some case, for example bind does not like it
        err = self._impl_valid_unicode(value)
        if err:
            return err
        if value.count('.') != 3:
            return ValueError()
        for val in value.split('.'):
            if val.startswith("0") and len(val) > 1:
                return ValueError()
        # 'standard' validation
        try:
            IP('{0}/32'.format(value))
        except ValueError:
            return ValueError()

    def _second_level_validation(self, value, warnings_only):
        ip = IP('{0}/32'.format(value))
        if not self._get_extra('_allow_reserved') and ip.iptype() == 'RESERVED':
            if warnings_only:
                msg = _("shouldn't in reserved class")
            else:
                msg = _("mustn't be in reserved class")
            return ValueError(msg)
        if self._get_extra('_private_only') and ip.iptype() != 'PRIVATE':
            if warnings_only:
                msg = _("should be in private class")
            else:
                msg = _("must be in private class")
            return ValueError(msg)

    def _cons_in_network(self, current_opt, opts, vals, warnings_only):
        if len(vals) != 3:
            return ConfigError(_('invalid len for vals'))
        if None in vals:
            return
        ip, network, netmask = vals
        if IP(ip) not in IP('{0}/{1}'.format(network, netmask)):
            msg = _('{4} is not in network {0}/{1} ({2}/{3})')
            return ValueError(msg.format(network, netmask,
                                         opts[1].impl_getname(), opts[2].impl_getname(), ip))
        # test if ip is not network/broadcast IP
        return opts[2]._cons_ip_netmask(current_opt, (opts[2], opts[0]), (netmask, ip), warnings_only)
