# -*- coding: utf-8 -*-
# Copyright (C) 2017 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
from IPy import IP

from ..setting import undefined
from ..i18n import _
from .option import Option


class NetworkOption(Option):
    "represents the choice of a network"
    __slots__ = tuple()
    _display_name = _('network address')

    def _validate(self, value, context=undefined, current_opt=undefined):
        err = self._impl_valid_unicode(value)
        if err:
            return err
        if value.count('.') != 3:
            return ValueError()
        for val in value.split('.'):
            if val.startswith("0") and len(val) > 1:
                return ValueError()
        try:
            IP(value)
        except ValueError:
            return ValueError()

    def _second_level_validation(self, value, warnings_only):
        ip = IP(value)
        if ip.iptype() == 'RESERVED':
            if warnings_only:
                msg = _("shouldn't be in reserved class")
            else:
                msg = _("mustn't be in reserved class")
            return ValueError(msg)
