# -*- coding: utf-8 -*-
# Copyright (C) 2017 Team tiramisu (see AUTHORS for all contributors)
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# The original `Config` design model is unproudly borrowed from
# the rough pypy's guys: http://codespeak.net/svn/pypy/dist/pypy/config/
# the whole pypy projet is under MIT licence
# ____________________________________________________________
import sys

from ..setting import undefined
from ..i18n import _
from .option import Option


class StrOption(Option):
    "represents the choice of a string"
    __slots__ = tuple()
    _display_name = _('string')

    def _validate(self, value, context=undefined, current_opt=undefined):
        if not isinstance(value, str):
            return ValueError()


if sys.version_info[0] >= 3:  # pragma: no cover
    #UnicodeOption is same as StrOption in python 3+
    class UnicodeOption(StrOption):
        __slots__ = tuple()
        pass
else:
    class UnicodeOption(Option):
        "represents the choice of a unicode string"
        __slots__ = tuple()
        _empty = u''
        _display_name = _('unicode string')

        def _validate(self, value, context=undefined, current_opt=undefined):
            if not isinstance(value, unicode):
                return ValueError()
